import java.util.Random;
import java.util.concurrent.Semaphore;

/**
 * Auf ein Subsystem, hier method1, method2 und method3, kann durch eine
 * gemeinsame Methode callMethode() zugegriffen werden. Das Subsystem enthlt 3
 * Objekte, welche zur Synchronisation verwendet werden. Wenn sich mehrere
 * Threads darin befinden, kann dies zu einem Deadlock fhren. Wenn am die
 * Methode callMethode mit Hilfe einer Semaphore jeweils nur zwei Threads in das
 * Subsystem einteten lsst, passiert dies nicht.
 */
public class SemaphoreDemo {
	
	// Beschrnkt die Anzahl Threads in den Methoden methode1,2,3
	private Semaphore s = new Semaphore(2);
	
	private Object lock1 = new Object();
	private Object lock2 = new Object();
	private Object lock3 = new Object();

	/**
	 * Zugriff auf die Methoden methode1,2,3
	 */
	private void callMethode(int id) {
		try {
			sleep(new Random().nextInt(100));
			/**/ s.acquire(); /**/ // = s.P()
			if (id==1) method1(1);
			if (id==2) method2(1);
			if (id==3) method3(1);
			/**/ s.release(); /**/ // = s.V()
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	private void method1(int n) {
		synchronized (lock1) {
			if (n==0) return;
			else {
				sleep(slpT);
				method2(n-1);
			}
		}
	}
	
	private void method2(int n) {
		synchronized (lock2) {
			if (n==0) return;
			else {
				sleep(slpT);
				method3(n-1);
			}
		}
	}
	
	private void method3(int n) {
		synchronized (lock3) {
			if (n==0) return;
			else {
				sleep(slpT);
				method1(n-1);
			}
		}
	}

	/**
	 * Erstelle 3 neue Threads und lasse sie jeweils eine der Methoden
	 * method1,2,3 aufrufen.
	 */
	public static void main(String[] args) {
		final SemaphoreDemo demo = new SemaphoreDemo();

		new Thread() {
			public void run() {
				while (true) {
					demo.callMethode(1);
					System.out.println("Thread 1 hat Methode 1 aufgerufen");
				}
			}
		}.start();
		
		new Thread() {
			public void run() {
				while (true) {
				demo.callMethode(2);
				System.out.println("Thread 2 hat Methode 2 aufgerufen");
				}
			}
		}.start();
		
		new Thread() {
			public void run() {
				while (true) {
				demo.callMethode(3);
				System.out.println("Thread 3 hat Methode 3 aufgerufen");
				}
			}
		}.start();
	}
	
	// Wie lange die Methoden method1,2,3 arbeiten mssen
	private static long slpT = 5;
	
	private static void sleep(long ms) {
		try {
			Thread.sleep(ms);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}
}
